from __future__ import annotations

from typing import Tuple

from core.constants import UnitProperty, CellValue, UnitClass
from core.constants.UnitProperty import attackUnitProperties
from core.logic.Command import Command, WORLD_PRIORITY, WORLD_MAX_WIDTH
from core.logic.Logic import Logic
from core.state import Unit
from tools.vector import vectorDistI


class AttackUnitCommand(Command):

    def __init__(self, cell: Tuple[int, int], targetCell: Tuple[int, int]):
        self.__cell = cell
        self.__targetCell = targetCell
        self.__damage = 0

    def priority(self) -> int:
        return WORLD_PRIORITY + self.__cell[0] + self.__cell[1] * WORLD_MAX_WIDTH

    def check(self, logic: Logic) -> bool:
        if self.__cell == self.__targetCell:
            return False

        world = logic.world
        if not world.contains(self.__cell):
            return False
        if not world.contains(self.__targetCell):
            return False

        unit = world.units.getUnit(self.__cell)
        if unit is None:
            return False
        if unit.playerId != logic.state.playerId:
            return False
        if not unit.hasProperty(UnitProperty.ACTION_POINTS):
            return False
        actionPoints = unit.getProperty(UnitProperty.ACTION_POINTS)
        if actionPoints == 0:
            return False

        targetUnit = world.units.getUnit(self.__targetCell)
        if targetUnit is None:
            return False
        if targetUnit.playerId == logic.state.playerId:
            return False
        if not targetUnit.hasProperty(UnitProperty.LIFE_POINTS):
            return False

        self.__damage = 0
        dist = vectorDistI(self.__cell, self.__targetCell)
        for attack, properties in attackUnitProperties.items():
            range = unit.getProperty(properties["range"], 1)
            if dist > range or (dist == 1 and range > 1):
                continue
            damage = unit.getProperty(attack, 0)
            damage -= targetUnit.getProperty(properties["defense"], 0)
            if damage > self.__damage:
                self.__damage = damage
        if self.__damage <= 0:
            return False

        return True

    def execute(self, logic: Logic):
        state = logic.state
        unitsLayer = logic.world.units

        unit = unitsLayer.getUnit(self.__cell)
        if unit is None:
            raise RuntimeError("No unit")
        actionPoints = unit.getProperty(UnitProperty.ACTION_POINTS)
        unit.setProperty(UnitProperty.ACTION_POINTS, actionPoints - 1)
        state.notifyUnitChanged(self.__cell, unit)

        targetUnit = unitsLayer.getUnit(self.__targetCell)
        if targetUnit is None:
            raise RuntimeError("No target unit")
        targetLifePoints = targetUnit.getProperty(UnitProperty.LIFE_POINTS)
        targetLifePoints = targetLifePoints - self.__damage
        if targetLifePoints > 0:
            targetUnit.setProperty(UnitProperty.LIFE_POINTS, targetLifePoints)
            state.notifyUnitChanged(self.__targetCell, targetUnit)
        elif targetUnit.unitClass == UnitClass.KNIGHT:
            targetUnit = Unit(UnitClass.SWORDSMAN, targetUnit.playerId)
            unitsLayer.setUnit(self.__targetCell, CellValue.UNITS_UNIT, targetUnit)
            unitsLayer.notifyCellChanged(self.__targetCell)
            state.notifyUnitChanged(self.__targetCell, targetUnit)
        else:
            unitsLayer.setUnit(self.__targetCell, CellValue.NONE, None)
            unitsLayer.notifyCellChanged(self.__targetCell)
            state.notifyUnitDied(self.__targetCell)
        state.notifyUnitDamaged(self.__targetCell, self.__damage)

