from __future__ import annotations

from abc import abstractmethod
from typing import Tuple, Optional, Union, TYPE_CHECKING

from pygame.rect import Rect
from pygame.surface import Surface

from core import Listenable
from core.constants import UnitClass
from tools.vector import vectorMulI, vectorAddI, vectorSubI, vectorMulF, vectorFtoI
from .IComponentListener import IComponentListener
from ..IUIEventHandler import IUIEventHandler
from ..theme.Theme import Theme

if TYPE_CHECKING:
    from ..Mouse import Mouse
    from ..mode.GameMode import GameMode


class Component(Listenable[IComponentListener]):
    def __init__(self, theme: Theme,
                 size: Optional[Tuple[int, int]] = None,
                 colorKey: Optional[Tuple[int, int, int]] = (0, 0, 0)):
        super().__init__()
        self.__theme = theme
        if size is None:
            self.__area = Rect(0, 0, theme.viewSize[0], theme.viewSize[1])
        else:
            self.__area = Rect(0, 0, size[0], size[1])
        self.__colorKey = colorKey

    @property
    def theme(self) -> Theme:
        return self.__theme

    def contains(self, pixel: Tuple[int, int]) -> bool:
        return self.__area.collidepoint(pixel) != 0

    @property
    def colorKey(self) -> Optional[Tuple[int, int, int]]:
        return self.__colorKey

    @property
    def area(self) -> Rect:
        return self.__area.copy()

    @property
    def topLeft(self) -> Tuple[int, int]:
        return int(self.__area.left), int(self.__area.top)

    @property
    def topRight(self) -> Tuple[int, int]:
        return int(self.__area.right), int(self.__area.top)

    @property
    def bottomLeft(self) -> Tuple[int, int]:
        return int(self.__area.left), int(self.__area.bottom)

    @property
    def bottomRight(self) -> Tuple[int, int]:
        return int(self.__area.right), int(self.__area.bottom)

    @property
    def size(self) -> Tuple[int, int]:
        return int(self.__area.width), int(self.__area.height)

    def resize(self, size: Tuple[int, int]):
        self.__area.size = size
    
    # Anchors

    def moveRelativeTo(self,
                       anchor: str,
                       other: Optional[Component] = None,
                       otherAnchor: str = "center",
                       borderX: Optional[int] = None,
                       borderY: Optional[int] = None,
                       borderSize: Optional[int] = None):
        if other is None:
            otherPos = (0, 0)
            otherSize = self.theme.viewSize
        else:
            otherPos = other.topLeft
            otherSize = other.size
        if borderSize is None:
            borderSize = self.theme.frameBorderSize

        assert otherAnchor in relativeAnchors, f"Invalid anchor {otherAnchor}"
        otherShift = vectorMulF(relativeAnchors[otherAnchor], otherSize)
        pos = vectorAddI(otherPos, vectorFtoI(otherShift))

        assert anchor in relativeAnchors, f"Invalid anchor {anchor}"
        shift = vectorMulF(relativeAnchors[anchor], self.size)
        pos = vectorSubI(pos, vectorFtoI(shift))

        defaultBorder = defaultBorderShifts[anchor][otherAnchor]
        defaultBorder = vectorMulI(defaultBorder, (borderSize, borderSize))
        if borderX is None:
            borderX = defaultBorder[0]
        if borderY is None:
            borderY = defaultBorder[1]
        pos = vectorAddI(pos, (borderX, borderY))

        self.moveTo(pos)

    def moveTo(self, topLeft: Tuple[int, int]):
        self.__area.update(topLeft, self.size)

    def shiftBy(self, shift: Tuple[int, int]):
        self.__area.move_ip(shift)

    # Component

    def dispose(self):
        pass

    def needRefresh(self) -> bool:
        return True

    @abstractmethod
    def render(self, surface: Surface):
        raise NotImplementedError()

    def findMouseFocus(self, mouse: Mouse) -> Optional[Component]:
        if not isinstance(self, IUIEventHandler):
            return None
        if not self.contains(mouse.pixel):
            return None
        return self

    # Component Listener

    def notifyWorldCellClicked(self, cell: Tuple[int, int], mouse: Mouse):
        for listener in self.listeners:
            listener.worldCellClicked(cell, mouse)

    def notifyWorldCellEntered(self, cell: Tuple[int, int], mouse: Mouse, dragging: bool):
        for listener in self.listeners:
            listener.worldCellEntered(cell, mouse, dragging)

    def notifyViewChanged(self, view: Tuple[int, int]):
        for listener in self.listeners:
            listener.viewChanged(view)

    # Game mode

    def notifySetGameModeRequested(self, gameMode: Optional[GameMode]):
        for listener in self.listeners:
            listener.setGameModeRequested(gameMode)

    def notifySetOverlayModeRequested(self, gameMode: Optional[GameMode]):
        for listener in self.listeners:
            listener.setOverlayModeRequested(gameMode)

    def notifyShowMessageRequested(self, message: str, portrait: str):
        for listener in self.listeners:
            listener.showMessageRequested(message, portrait)

    # Edition

    def notifyMainBrushSelected(self, layerName: str, value: Union[int, str], unitClass: Optional[UnitClass]):
        for listener in self.listeners:
            listener.mainBrushSelected(layerName, value, unitClass)

    def notifySecondaryBrushSelected(self, layerName: str, value: Union[int, str], unitClass: Optional[UnitClass]):
        for listener in self.listeners:
            listener.secondaryBrushSelected(layerName, value, unitClass)


relativeAnchors = {
    None: (0, 0),
    "topLeft": (0, 0),
    "top": (0.5, 0),
    "topRight": (1, 0),
    "left": (0, 0.5),
    "center": (0.5, 0.5),
    "right": (1, 0.5),
    "bottomLeft": (0, 1),
    "bottom": (0.5, 1),
    "bottomRight": (1, 1)
}

defaultBorderShifts = {
    "topLeft": {
        "topLeft": (1, 1),
        "top": (0, 1),
        "topRight": (1, 0),
        "left": (1, 0),
        "center": (0, 0),
        "right": (1, 0),
        "bottomLeft": (0, 1),
        "bottom": (0, 1),
        "bottomRight": (1, 1)
    },
    "top": {
        "topLeft": (0, 1),
        "top": (0, 1),
        "topRight": (0, 1),
        "left": (0, 0),
        "center": (0, 0),
        "right": (0, 0),
        "bottomLeft": (0, 1),
        "bottom": (0, 1),
        "bottomRight": (0, 1)
    },
    "topRight": {
        "topLeft": (-1, 0),
        "top": (0, 1),
        "topRight": (-1, 1),
        "left": (-1, 0),
        "center": (0, 0),
        "right": (-1, 0),
        "bottomLeft": (-1, 1),
        "bottom": (0, 1),
        "bottomRight": (0, 1)
    },
    "left": {
        "topLeft": (1, 0),
        "top": (0, 0),
        "topRight": (1, 0),
        "left": (1, 0),
        "center": (0, 0),
        "right": (1, 0),
        "bottomLeft": (1, 0),
        "bottom": (0, 0),
        "bottomRight": (1, 0)
    },
    "center": {
        "topLeft": (0, 0),
        "top": (0, 0),
        "topRight": (0, 0),
        "left": (0, 0),
        "center": (0, 0),
        "right": (0, 0),
        "bottomLeft": (0, 0),
        "bottom": (0, 0),
        "bottomRight": (0, 0)
    },
    "right": {
        "topLeft": (-1, 0),
        "top": (0, 0),
        "topRight": (-1, 0),
        "left": (-1, 0),
        "center": (0, 0),
        "right": (-1, 0),
        "bottomLeft": (-1, 0),
        "bottom": (0, 0),
        "bottomRight": (-1, 0)
    },
    "bottomLeft": {
        "topLeft": (0, -1),
        "top": (0, -1),
        "topRight": (1, -1),
        "left": (1, 0),
        "center": (0, 0),
        "right": (1, 0),
        "bottomLeft": (1, -1),
        "bottom": (0, -1),
        "bottomRight": (1, 0)
    },
    "bottom": {
        "topLeft": (0, -1),
        "top": (0, -1),
        "topRight": (0, -1),
        "left": (0, 0),
        "center": (0, 0),
        "right": (0, 0),
        "bottomLeft": (0, -1),
        "bottom": (0, -1),
        "bottomRight": (0, -1)
    },
    "bottomRight": {
        "topLeft": (-1, -1),
        "top": (0, -1),
        "topRight": (0, -1),
        "left": (-1, 0),
        "center": (0, 0),
        "right": (-1, 0),
        "bottomLeft": (-1, 0),
        "bottom": (0, -1),
        "bottomRight": (-1, -1)
    },
}