from typing import Optional, Tuple

from pygame.surface import Surface

from core.logic import DistanceMap
from core.state import Layer, GameState
from tools.vector import vectorSubI, vectorSubDiv2I
from ui.layer.SelectionValue import SelectionValue
from .LayerComponent import LayerComponent
from .LayerRenderer import LayerRenderer
from ..text.token import TextStyle
from ...theme.Theme import Theme


class SelectionComponent(LayerComponent):
    def __init__(self, theme: Theme, state: GameState, layer: Layer):
        super().__init__(theme, layer, "selection")
        self.__state = state
        self.__selectionLayer = layer
        self.__damageCell: Optional[Tuple[int, int]] = None
        self.__damage = 0
        self.__damageShift = 0

        # Debug / Illustrate
        self.__textStyle = TextStyle("small", "white")
        self.__distanceMap: Optional[DistanceMap] = None
        self.__showMode = 0

    def addDamage(self, cell: Tuple[int, int], damage: int):
        self.__damage = damage
        self.__damageCell = cell
        self.__damageShift = 0

    def render(self, surface: Surface):
        super().render(surface)
        tileset = self.tileset.surface
        tileSize = self.tileset.tileSize
        tilesRects = self.tileset.getTilesRects()

        renderer = self.createRenderer(surface)
        cellsSlice = renderer.cellsSlice
        cells = self.__selectionLayer.cells[cellsSlice]

        valid = cells == SelectionValue.SELECTED
        currentPlayerId = self.__state.playerId
        for dest, value, cell in renderer.coords(valid):
            rect = tilesRects[value][currentPlayerId]
            shift = vectorSubDiv2I(rect.size, tileSize)
            dest = vectorSubI(dest, shift)
            surface.blit(tileset, dest, rect)

        valid = cells != SelectionValue.NONE
        valid &= cells != SelectionValue.SELECTED
        for dest, value, cell in renderer.coords(valid):
            rects = tilesRects[value]
            if len(rects) == 1:
                rect = rects[0]
            else:
                rect = rects[currentPlayerId]
            surface.blit(tileset, dest, rect)

        # Debug / Illustrate
        if self.__distanceMap is None:
            return
        if self.__showMode == 1:
            self.__showEdges(surface, renderer)
        elif self.__showMode == 2:
            self.__showMap(surface, renderer)

    # Debug / Illustrate

    def setDistanceMap(self, distanceMap: Optional[DistanceMap]):
        self.__distanceMap = distanceMap

    def rollShowMode(self):
        self.__showMode += 1
        if self.__showMode > 2:
            self.__showMode = 0
        self._needRefresh = True

    def __showMap(self, surface: Surface, renderer: LayerRenderer):
        if self.__distanceMap is None:
            return
        map = self.__distanceMap.map
        ax1, ay1, ax2, ay2 = self.__distanceMap.area
        cellMinX, cellMinY = renderer.topLeft
        ax1 -= cellMinX
        ay1 -= cellMinY
        ax2 -= cellMinX
        ay2 -= cellMinY

        textStyle = TextStyle("small", "white")
        tileWidth, tileHeight = self.theme.getTileset("ground").tileSize
        for dest, value, cell in renderer.coords():
            x, y = cell
            if x < ax1 or y < ay1 or x >= ax2 or y >= ay2:
                continue
            mapX, mapY = x - ax1, y - ay1
            x, y = dest
            value = map[mapX, mapY]
            if value < 100:
                textSurface = textStyle.render(self.theme, value)
                textSurface.set_colorkey((0, 0, 0))
                surface.blit(textSurface, (
                    x + (tileWidth - textSurface.get_width()) // 2,
                    y + (tileHeight - textSurface.get_height()) // 2
                ))

    def __showEdges(self, surface: Surface, renderer: LayerRenderer):
        if self.__distanceMap is None:
            return

        edges = self.__distanceMap.edges
        ax1, ay1, ax2, ay2 = self.__distanceMap.area
        cellMinX, cellMinY = renderer.topLeft
        ax1 -= cellMinX
        ay1 -= cellMinY
        ax2 -= cellMinX
        ay2 -= cellMinY

        textStyle = TextStyle("small", "white")
        tileWidth, tileHeight = self.theme.getTileset("ground").tileSize
        for dest, value, cell in renderer.coords():
            x, y = cell
            if x < ax1 or y < ay1 or x >= ax2 or y >= ay2:
                continue
            mapX, mapY = x - ax1, y - ay1
            x, y = dest
            # Top Left
            value = edges[mapX, mapY, 0]
            if value < 10:
                textSurface = textStyle.render(self.theme, value)
                textSurface.set_colorkey((0, 0, 0))
                surface.blit(textSurface, (
                    x - 1,
                    y
                ))
            # Left
            value = edges[mapX, mapY, 1]
            if value < 10:
                textSurface = textStyle.render(self.theme, value)
                textSurface.set_colorkey((0, 0, 0))
                surface.blit(textSurface, (
                    x - 1,
                    y + (tileHeight - textSurface.get_height()) // 2
                ))
            # Bottom Left
            value = edges[mapX, mapY, 2]
            if value < 10:
                textSurface = textStyle.render(self.theme, value)
                textSurface.set_colorkey((0, 0, 0))
                surface.blit(textSurface, (
                    x - 1,
                    y + tileHeight - textSurface.get_height()
                ))
            # Top
            value = edges[mapX, mapY, 3]
            if value < 10:
                textSurface = textStyle.render(self.theme, value)
                textSurface.set_colorkey((0, 0, 0))
                surface.blit(textSurface, (
                    x + (tileWidth - textSurface.get_width()) // 2,
                    y
                ))
            # Bottom
            value = edges[mapX, mapY, 4]
            if value < 10:
                textSurface = textStyle.render(self.theme, value)
                textSurface.set_colorkey((0, 0, 0))
                surface.blit(textSurface, (
                    x + (tileWidth - textSurface.get_width()) // 2,
                    y + tileHeight - textSurface.get_height()
                ))
            # Top Right
            value = edges[mapX, mapY, 5]
            if value < 10:
                textSurface = textStyle.render(self.theme, value)
                textSurface.set_colorkey((0, 0, 0))
                surface.blit(textSurface, (
                    x + tileWidth - textSurface.get_width(),
                    y
                ))
            # Right
            value = edges[mapX, mapY, 6]
            if value < 10:
                textSurface = textStyle.render(self.theme, value)
                textSurface.set_colorkey((0, 0, 0))
                surface.blit(textSurface, (
                    x + tileWidth - textSurface.get_width(),
                    y + (tileHeight - textSurface.get_height()) // 2
                ))
            # Bottom Right
            value = edges[mapX, mapY, 7]
            if value < 10:
                textSurface = textStyle.render(self.theme, value)
                textSurface.set_colorkey((0, 0, 0))
                surface.blit(textSurface, (
                    x + tileWidth - textSurface.get_width(),
                    y + tileHeight - textSurface.get_height()
                ))


